# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""consists of DVCSteps to embedd files and save them as for example as csv."""

# Standard library imports
import os
import re
from io import StringIO
from logging import getLogger
from pathlib import Path
from typing import Optional, TypedDict

from markdown import Markdown
from pandera.typing import DataFrame
from tqdm.auto import tqdm

from wurzel.datacontract import MarkdownDataContract
from wurzel.exceptions import EmbeddingAPIException, StepFailed
from wurzel.step import TypedStep
from wurzel.steps.embedding.huggingface import HuggingFaceInferenceAPIEmbeddings, PrefixedAPIEmbeddings
from wurzel.steps.splitter import SimpleSplitterStep
from wurzel.utils.semantic_splitter import SemanticSplitter

from .data import EmbeddingResult

# Local application/library specific imports
from .settings import EmbeddingSettings

log = getLogger(__name__)


class Embedded(TypedDict):
    """dict definition of a embedded document."""

    text: str
    url: str
    vector: list[float]


class EmbeddingStep(
    SimpleSplitterStep,
    TypedStep[EmbeddingSettings, list[MarkdownDataContract], DataFrame[EmbeddingResult]],
):
    """Step for consuming list[MarkdownDataContract]
    and returning DataFrame[EmbeddingResult].
    """

    embedding: HuggingFaceInferenceAPIEmbeddings
    n_jobs: int
    markdown: Markdown
    stopwords: list[str]

    def __init__(self) -> None:
        super().__init__()
        self.settings = EmbeddingSettings()
        self.embedding = self._select_embedding(self.settings.API)
        self.n_jobs = max(1, (os.cpu_count() or 0) - 1)
        # Inject net output_format into 3rd party library Markdown
        Markdown.output_formats["plain"] = self.__md_to_plain  # type: ignore[index]
        self.markdown = Markdown(output_format="plain")  # type: ignore[arg-type]
        self.markdown.stripTopLevelTags = False
        self.settingstopwords = self._load_stopwords(self.settings.STEPWORDS_PATH)
        self.splitter = SemanticSplitter(
            token_limit=self.settings.TOKEN_COUNT_MAX,
            token_limit_buffer=self.settings.TOKEN_COUNT_BUFFER,
            token_limit_min=self.settings.TOKEN_COUNT_MIN,
        )

    @staticmethod
    def _load_stopwords(path: Path) -> list[str]:
        with path.open(encoding="utf-8") as f:
            stopwords = [w.strip() for w in f.readlines() if not w.startswith(";")]
        return stopwords

    @staticmethod
    def _select_embedding(*args, **kwargs) -> HuggingFaceInferenceAPIEmbeddings:
        """Selects the embedding model to be used for generating embeddings.

        Returns
        -------
        Embeddings
            An instance of the Embeddings class.

        """
        return PrefixedAPIEmbeddings(*args, **kwargs)

    def run(self, inpt: list[MarkdownDataContract]) -> DataFrame[EmbeddingResult]:
        """Executes the embedding step by processing input markdown files, generating embeddings,
        and saving them to a CSV file.
        """
        if len(inpt) == 0:
            log.info("Got empty result in Embedding - Skipping")
            return DataFrame[EmbeddingResult]([])
        splitted_md_rows = self._split_markdown(inpt)
        rows = []
        failed = 0
        for row in tqdm(splitted_md_rows, desc="Calculate Embeddings"):
            try:
                rows.append(self._get_embedding(row))
            except EmbeddingAPIException as err:
                log.warning(
                    f"Skipped because EmbeddingAPIException: {err.message}",
                    extra={"markdown": str(row)},
                )
                failed += 1
        if failed:
            log.warning(f"{failed}/{len(splitted_md_rows)} got skipped")
        if failed == len(splitted_md_rows):
            raise StepFailed(f"all {len(splitted_md_rows)} embeddings got skipped")
        return DataFrame[EmbeddingResult](DataFrame[EmbeddingResult](rows))

    def get_embedding_input_from_document(self, doc: MarkdownDataContract) -> str:
        """Clean the document such that it can be used as input to the embedding model.

        Parameters
        ----------
        doc : MarkdownDataContract
            The document containing the page content in Markdown format.

        Returns
        -------
        str
            Cleaned text that can be used as input to the embedding model.

        """
        plain_text = self.markdown.convert(doc.md)
        plain_text = self._replace_link(plain_text)

        return plain_text

    def _get_embedding(self, doc: MarkdownDataContract) -> Embedded:
        """Generates an embedding for a given text and context.

        Parameters
        ----------
        d : dict
            A dictionary containing the text and context for which to generate the embedding.

        Returns
        -------
        dict
            A dictionary containing the original text, its embedding, and the source URL.

        """
        context = self.get_simple_context(doc.keywords)
        plain_text = self.get_embedding_input_from_document(doc)
        vector = self.embedding.embed_query(plain_text)
        return {"text": doc.md, "vector": vector, "url": doc.url, "keywords": context}

    def is_stopword(self, word: str) -> bool:
        """Stopword Detection Function."""
        return word.lower() in self.settingstopwords

    @classmethod
    def whitespace_word_tokenizer(cls, text: str) -> list[str]:
        """Simple Regex based whitespace word tokenizer."""
        return [x for x in re.split(r"([.,!?]+)?\s+", text) if x]

    def get_simple_context(self, text):
        """Simple function to create a context from a text."""
        tokens = self.whitespace_word_tokenizer(text)
        filtered_tokens = [token for token in tokens if not self.is_stopword(token)]
        return " ".join(filtered_tokens)

    @classmethod
    def __md_to_plain(cls, element, stream: Optional[StringIO] = None):
        """Converts a markdown element into plain text.

        Parameters
        ----------
        element : Element
            The markdown element to convert.
        stream : StringIO, optional
            The stream to which the plain text is written. If None, a new stream is created.

        Returns
        -------
        str
            The plain text representation of the markdown element.

        """
        if stream is None:
            stream = StringIO()
        if element.text:
            stream.write(element.text)
        for sub in element:
            cls.__md_to_plain(sub, stream)
        if element.tail:
            stream.write(element.tail)
        return stream.getvalue()

    @classmethod
    def _replace_link(cls, text: str):
        """Replaces URLs in the text with a placeholder.

        Parameters
        ----------
        text : str
            The text in which URLs will be replaced.

        Returns
        -------
        str
            The text with URLs replaced by 'LINK'.

        """
        # Extract URL from a string
        url_extract_pattern = (
            "https?:\\/\\/(?:www\\.)?[-a-zA-Z0-9@:%._\\+~#=]{1,256}\\.[a-zA-Z0-9()]{1,6}\\b(?:[-a-zA-Z0-9()@:%_\\+.~#?&\\/=]*)"  # pylint: disable=line-too-long
        )
        links = sorted(re.findall(url_extract_pattern, text), key=len, reverse=True)
        for link in links:
            text = text.replace(link, "LINK")
        return text



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import logging.config
import os
import sys
from collections.abc import Mapping
from typing import Any, Optional

from asgi_correlation_id import correlation_id

log = logging.getLogger(__name__)


# pylint: disable-next=too-many-positional-arguments
def warnings_to_logger(message: str, category: str, filename: str, lineno: str, file=None, line=None):
    """Replaces the default `warnings.showwarning` function to log warnings using the Python logging framework.

    This function captures warning messages and logs them to a logger associated with the module where the warning originated.
    If the module cannot be determined, it uses the filename as the logger name.

        message (str): The warning message to be logged.
        category (str): The category of the warning (e.g., `UserWarning`, `DeprecationWarning`).
        filename (str): The name of the file where the warning was triggered.
        lineno (str): The line number in the file where the warning was triggered.
        file (Optional[IO], optional): Not used. Included for compatibility with `warnings.showwarning`. Defaults to None.
        line (Optional[str], optional): Not used. Included for compatibility with `warnings.showwarning`. Defaults to None.

    """
    # pylint: disable=unused-argument
    for module_name, module in sys.modules.items():
        module_path = getattr(module, "__file__", None)
        if module_path and os.path.abspath(module_path) == os.path.abspath(filename):
            break
    else:
        module_name = os.path.splitext(os.path.split(filename)[1])[0]
    logger = logging.getLogger(module_name)
    extra = {
        "warnings.category": category,
        "warnings.filename": filename,
        "warnings.lineno": lineno,
    }
    logger.warning(message, extra=extra)


def _make_dict_serializable(item: Any):
    secret_words = ["password", "key", "secret"]
    match item:
        case dict():
            new_dict = {}
            for k, v in item.items():
                # If value is None remove item
                if v is None:
                    continue
                key = k if isinstance(k, str) else repr(k)
                # Could also use SecretStr but changes usage of Settings object
                if any(keyword in key.lower() for keyword in secret_words):
                    new_dict[key] = "****"
                else:
                    new_dict[key] = _make_dict_serializable(v)
            return new_dict
        case str() | int() | float():
            return item
        case list() | set():
            return [_make_dict_serializable(i) for i in item]
        case _:
            return repr(item)


class JsonFormatter(logging.Formatter):
    """Custom formatter for structured logging."""

    key_blacklist = [
        "msg",
        "message",
        "args",
        "created",
        "msecs",
        "relativeCreated",
        "levelno",
        "filename",
        "color_message",
    ]

    def __init__(
        self,
        datefmt: Optional[str] = "%Y-%m-%dT%H:%M:%S%z",
        reduced: Optional[list[str]] = None,
        indent: Optional[str] = None,
        *,
        defaults: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """Create a new Formatter.

        Args:
            datefmt (str, optional): Used in @timestamp. Defaults to "%Y-%m-%dT%H:%M:%S%z".
            reduced (Optional[List[str]], optional): List of loglevels to reduce output by. Defaults to None.
                Reduced output removes filename, thread and process information
            indent (Optional[str], optional): indent for json dumps. Defaults to None.

        """
        super().__init__(None, datefmt, defaults=defaults)
        self.indent = indent
        self.reduced_levels = [logging.getLevelNamesMapping().get(level) for level in reduced or []]

    def _get_output_dict(self, record: logging.LogRecord) -> dict[str, Any]:
        data = {k: v for k, v in record.__dict__.items() if k not in self.key_blacklist and v is not None}
        logger_name = f"{data.pop('module')}.{data.pop('name')}"
        func_name = data.pop("funcName")
        if func_name != "<module>":
            logger_name = logger_name + "." + func_name
        output = {
            "level": data.pop("levelname"),
            "message": record.getMessage(),
            "logger_name": logger_name,
            "file": f"{data.pop('pathname')}:{data.pop('lineno')}",
            "extra": {},
            "@timestamp": self.formatTime(record, self.datefmt),
            "process": f"{data.pop('processName')}({data.pop('process')})",
            "thread": f"{data.pop('threadName')}({data.pop('thread')})",
        }
        if all(key in data for key in ["warnings.category", "warnings.filename", "warnings.lineno"]):
            output["file"] = f"{data.pop('warnings.filename')}:{data.pop('warnings.lineno')}"
        if data:
            output["extra"] = _make_dict_serializable(data)
        cor_id = correlation_id.get()
        if cor_id is not None:
            output["correlationId"] = cor_id
        if self.reduced_levels and record.levelno in self.reduced_levels:
            del output["process"]
            del output["logger_name"]
            del output["thread"]
        if not output["extra"]:
            del output["extra"]
        return output

    def format(self, record: logging.LogRecord) -> str:
        super().format(record)
        output = self._get_output_dict(record)
        return json.dumps(output, default=repr, indent=self.indent)


def get_logging_dict_config(level) -> dict[str, str]:
    """Generate a logging.config.dictConfig compatible dict.

    Returns:
        dict: logging.config.dictConfig

    """
    default_formatter = {
        "json_formatter": {
            "()": "wurzel.utils.logging.JsonFormatter",
            "reduced": ["INFO"],
        }
    }
    default_handler = {
        "default": {
            "level": level,
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
            "formatter": "json_formatter",
        },
    }
    logger_template = {"level": level, "handlers": ["default"], "propagate": False}
    return {
        "version": 1,
        "root": {},
        "disable_existing_loggers": False,
        "formatters": {**default_formatter},
        "handlers": {**default_handler},
        "loggers": {
            "root": {**logger_template},
            "": {**logger_template},
            "uvicorn.error": {**logger_template},
            "uvicorn.access": {
                "level": "WARNING",
                "handlers": ["default"],
                "propagate": False,
            },
            "gunicorn.access": {"propagate": True},
            "gunicorn.error": {"propagate": True},
            "transaction": {**logger_template},
        },
    }



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import json
from typing import Optional

from pydantic import Field, validator
from qdrant_client.models import Distance

from wurzel.step.settings import Settings


# pylint: disable=duplicate-code
class QdrantSettings(Settings):
    """QdrantSettings is a configuration class for managing settings related to the Qdrant database.

    Attributes:
        DISTANCE (Distance): The distance metric to be used, default is Distance.DOT.
        URI (str): The URI for the Qdrant database, default is "http://localhost:6333".
        COLLECTION (str): The name of the collection in the Qdrant database.
        COLLECTION_HISTORY_LEN (int): The length of the collection history, default is 10.
        SEARCH_PARAMS (dict): Parameters for search operations, default is {"metric_type": "IP", "params": {}}.
        INDEX_PARAMS (dict): Parameters for index creation, default includes "index_type", "field_name", "distance", and "params".
        APIKEY (Optional[str]): The API key for authentication, default is an empty string.
        REPLICATION_FACTOR (int): The replication factor for the database, default is 3, must be greater than 0.
        BATCH_SIZE (int): The batch size for operations, default is 1024, must be greater than 0.

    Methods:
        parse_json(v):
            Validates and parses JSON strings into Python objects for SEARCH_PARAMS and INDEX_PARAMS.

    Args:
                v (Union[str, dict]): The input value, either a JSON string or a dictionary.

    Returns:
                dict: The parsed dictionary if the input is a JSON string, otherwise the input value.

    """

    DISTANCE: Distance = Distance.DOT
    URI: str = "http://localhost:6333"
    COLLECTION: str
    COLLECTION_HISTORY_LEN: int = 10
    SEARCH_PARAMS: dict = {"metric_type": "IP", "params": {}}
    INDEX_PARAMS: dict = {
        "index_type": "FLAT",
        "field_name": "vector",
        "distance": "Dot",
        "params": {},
    }
    APIKEY: Optional[str] = ""
    REPLICATION_FACTOR: int = Field(default=3, gt=0)
    BATCH_SIZE: int = Field(default=1024, gt=0)

    @validator("SEARCH_PARAMS", "INDEX_PARAMS", pre=True)
    @classmethod
    def parse_json(cls, v):
        """Validation for json."""
        if isinstance(v, str):
            return json.loads(v)
        return v



# SPDX-FileCopyrightText: 2024 Deutsche Telekom AG
#
# SPDX-License-Identifier: CC0-1.0

"""Note: Known Limitations with EasyOCR (`EasyOcrOptions`).

1. Table structure is often lost or misaligned in the OCR output.
2. Spelling inaccuracies are occasionally observed (e.g., "Verlängern" → "Verlängenng").
3. URLs are not parsed correctly (e.g., "www.telekom.de/agb" → "www telekom delagb").

While investigating EasyOCR issues and testing alternative OCR engines,
we observed that some documents produced distorted text with irregular whitespace.
This disrupts the natural sentence flow and significantly reduces readability.

Example:
"pra kti sche  i nform ati o nen zu  i h rer  fam i l y  card  basi c Li eber
  Tel ekom   Kunde, schön,   dass  Si e  si ch  f ür..."

Despite these limitations, we have decided to proceed with EasyOCR.

"""

from logging import getLogger

from bs4 import BeautifulSoup, Comment
from docling.datamodel.base_models import InputFormat
from docling.datamodel.pipeline_options import EasyOcrOptions, PdfPipelineOptions
from docling.document_converter import (
    DocumentConverter,
    PdfFormatOption,
)
from mistletoe import Document as MTDocument
from mistletoe import HTMLRenderer

from wurzel.datacontract.common import MarkdownDataContract
from wurzel.step.typed_step import TypedStep
from wurzel.utils.to_markdown.html2md import MD_RENDER_LOCK

from .settings import DoclingSettings

log = getLogger(__name__)


class CleanMarkdownRenderer(HTMLRenderer):
    """Custom Markdown renderer extending mistletoe's HTMLRenderer to clean up
    unwanted elements from Markdown input.
    """

    @staticmethod
    def render_html_block(token):
        """Render HTML block tokens and clean up unwanted elements.

        This method removes HTML comments and returns the cleaned HTML content.
        Remove comments like <!-- image -->
        """
        soup = BeautifulSoup(token.content, "html.parser")

        for comment in soup.find_all(string=lambda text: isinstance(text, Comment)):
            comment.extract()
        return soup.decode_contents().strip()


class DoclingStep(TypedStep[DoclingSettings, None, list[MarkdownDataContract]]):
    """Step to return local Markdown files with enhanced PDF extraction for German."""

    def __init__(self):
        super().__init__()
        self.converter = self.create_converter()

    def create_converter(self) -> DocumentConverter:
        """Create and configure the document converter for PDF and DOCX.

        Returns:
            DocumentConverter: Configured document converter.

        """
        pipeline_options = PdfPipelineOptions()
        ocr_options = EasyOcrOptions()
        pipeline_options.ocr_options = ocr_options

        return DocumentConverter(
            allowed_formats=self.settings.FORMATS,
            format_options={
                InputFormat.PDF: PdfFormatOption(
                    pipeline_options=pipeline_options,
                )
            },
        )

    @staticmethod
    def extract_keywords(md_text: str) -> str:
        """Cleans a Markdown string using mistletoe and extracts useful content.

        - Parses and renders the Markdown content into HTML using a custom HTML renderer
        - Removes unwanted HTML comments and escaped underscores
        - Extracts the first heading from the content (e.g., `<h1>` to `<h6>`)
        - Converts the cleaned HTML into plain text

        Args:
            md_text (str): The raw Markdown input string.

        """
        with MD_RENDER_LOCK, CleanMarkdownRenderer() as renderer:
            ast = MTDocument(md_text)
            cleaned = renderer.render(ast).replace("\n", "")
            soup = BeautifulSoup(cleaned, "html.parser")
            first_heading_tag = soup.find(["h1", "h2", "h3", "h4", "h5", "h6"])
            heading = first_heading_tag.get_text(strip=True) if first_heading_tag else ""

        return heading

    def run(self, inpt: None) -> list[MarkdownDataContract]:
        """Run the document extraction and conversion process for German PDFs.

        Args:
            inpt (None): Input parameter (not used).

        Returns:
            List[MarkdownDataContract]: List of converted Markdown contracts.

        """
        urls = self.settings.URLS
        contracts = []

        for url in urls:
            try:
                converted_contract = self.converter.convert(url)
                md = converted_contract.document.export_to_markdown(image_placeholder="")
                keyword = self.extract_keywords(md)
                contract_instance = {"md": md, "keywords": " ".join([self.settings.DEFAULT_KEYWORD, keyword]), "url": url}
                contracts.append(contract_instance)

            except (FileNotFoundError, OSError) as e:
                log.error(f"Failed to verify URL: {url}. Error: {e}")
                continue

        return contracts



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from wurzel.utils import HAS_DOCLING as _HAS_DOCLING

if _HAS_DOCLING:
    from .docling_step import DoclingStep  # noqa: F401
    from .settings import DoclingSettings  # noqa: F401
else:
    pass



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import re
from pathlib import Path
from typing import Annotated

from pydantic import Field, WrapValidator
from pydantic_core import Url

from wurzel.steps.splitter import SplitterSettings


class EmbeddingSettings(SplitterSettings):
    """EmbeddingSettings is a configuration class for embedding-related settings.

    Attributes:
        API (Url): The API endpoint for embedding operations.
        NORMALIZE (bool): A flag indicating whether to normalize embeddings. Defaults to False.
        BATCH_SIZE (int): The batch size for processing embeddings. Must be greater than 0. Defaults to 100.
        TOKEN_COUNT_MIN (int): The minimum token count for processing. Must be greater than 0. Defaults to 64.
        TOKEN_COUNT_MAX (int): The maximum token count for processing. Must be greater than 1. Defaults to 256.
        TOKEN_COUNT_BUFFER (int): The buffer size for token count. Must be greater than 0. Defaults to 32.
        STEPWORDS_PATH (Path): The file path to the stopwords file. Defaults to "data/german_stopwords_full.txt".
        N_JOBS (int): The number of parallel jobs to use. Must be greater than 0. Defaults to 1.
        PREFIX_MAP (dict[re.Pattern, str]): A mapping of regex patterns to string prefixes.
            This is validated and transformed using the `_wrap_validator_model_mapping` method.

    Methods:
        _wrap_validator_model_mapping(input_dict: dict[str, str], handler):
            A static method to wrap and validate the model mapping. It converts string regex keys
            in the input dictionary to compiled regex patterns and applies a handler function to the result.

    """

    @staticmethod
    def _wrap_validator_model_mapping(input_dict: dict[str, str], handler):
        new_dict = {}
        for regex, prefix in input_dict.items():
            if isinstance(regex, str):
                new_dict[re.compile(regex)] = prefix
            else:
                new_dict.update({regex: prefix})
        return handler(new_dict)

    API: Url
    NORMALIZE: bool = False
    BATCH_SIZE: int = Field(100, gt=0)
    TOKEN_COUNT_MIN: int = Field(64, gt=0)
    TOKEN_COUNT_MAX: int = Field(256, gt=1)
    TOKEN_COUNT_BUFFER: int = Field(32, gt=0)
    STEPWORDS_PATH: Path = Path("data/german_stopwords_full.txt")
    N_JOBS: int = Field(1, gt=0)
    PREFIX_MAP: Annotated[dict[re.Pattern, str], WrapValidator(_wrap_validator_model_mapping)] = Field(
        default={"e5-": "query : ", "DPR|dpr": ""}
    )



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""containing the DVCStep sending embedding data into milvus."""

from logging import getLogger

import pandas as pd
from pandera.typing import DataFrame
from pymilvus import CollectionSchema, DataType, FieldSchema
from pymilvus.exceptions import MilvusException
from pymilvus.milvus_client import MilvusClient
from pymilvus.milvus_client.index import IndexParams

from wurzel.exceptions import NoPreviousCollection, StepFailed
from wurzel.step import TypedStep
from wurzel.steps.embedding.data import EmbeddingResult

from .data import Result as MilvusResult
from .settings import MilvusSettings

log = getLogger(__name__)


class MilvusConnectorStep(TypedStep[MilvusSettings, DataFrame[EmbeddingResult], MilvusResult]):  # pragma: no cover
    """Milvus connector step. It consumes embedding csv files, creates a new schema and inserts the embeddings."""

    milvus_timeout: float = 20.0

    def __init__(self) -> None:
        super().__init__()
        # milvus stuff passed as environment
        # because we need to enject them into the DVC step during runtime,
        # not during DVC pipeline definition time
        uri = f"http://{self.settings.HOST}:{self.settings.PORT}"
        if not self.settings.PASSWORD or not self.settings.USER:
            log.warning("MILVUS_HOST, MILVUS_USER or MILVUS_PASSWORD for Milvus not provided. Thus running in non-credential Mode")
        self.client: MilvusClient = MilvusClient(
            uri=uri,
            user=self.settings.USER,
            password=self.settings.PASSWORD,
            timeout=self.milvus_timeout,
        )
        self.collection_index: IndexParams = IndexParams(**self.settings.INDEX_PARAMS)
        self.collection_history_len = self.settings.COLLECTION_HISTORY_LEN

        self.collection_prefix = self.settings.COLLECTION

    def __del__(self):
        if getattr(self, "client", None):
            self.client.close()

    def run(self, inpt: DataFrame[EmbeddingResult]) -> MilvusResult:
        self._insert_embeddings(inpt)
        try:
            old = self.__construct_last_collection_name()
        except NoPreviousCollection:
            old = ""
        self._retire_collection()
        return MilvusResult(new=self.__construct_current_collection_name(), old=old)

    def _insert_embeddings(self, data: pd.DataFrame):
        collection_name = self.__construct_next_collection_name()
        log.info(f"Creating milvus collection {collection_name}")
        collection_schema = CollectionSchema(
            fields=[
                FieldSchema(name="pk", dtype=DataType.INT64, is_primary=True, auto_id=True),
                FieldSchema(name="text", dtype=DataType.VARCHAR, max_length=3000),
                FieldSchema(
                    name="vector",
                    dtype=DataType.FLOAT_VECTOR,
                    dim=len(data["vector"].loc[0]),
                ),
                FieldSchema(name="url", dtype=DataType.VARCHAR, max_length=300),
            ],
            description="Collection for storing Milvus embeddings",
        )

        log.info("schema created")
        self.client.create_collection(collection_name=collection_name, schema=collection_schema)
        log.info("collection created")
        log.info(f"Inserting embedding {len(data)} into collection {collection_name}")
        result: dict = self.client.insert(collection_name=collection_name, data=data.to_dict("records"))
        if result["insert_count"] != len(data):
            raise StepFailed(
                f"Failed to insert df into collection '{collection_name}'.{result['insert_count']}/{len(data)} where successful"
            )
        log.info(f"Successfully inserted {len(data)} vectors into collection '{collection_name}'")
        self.client.create_index(collection_name=collection_name, index_params=self.collection_index)
        log.info(f"Successfully craeted index {self.collection_index} into collection '{collection_name}")
        self.client.load_collection(collection_name)
        log.info(f"Successfully loaded the collection {collection_name}' into collection '{collection_name}'")
        try:
            self.client.release_collection(self.__construct_last_collection_name())
        except NoPreviousCollection:
            pass
        self._update_alias(collection_name)

    def _retire_collection(self):
        collections_versioned: dict[int, str] = self._get_collection_versions()
        to_delete = sorted(collections_versioned.keys())[: -self.collection_history_len]
        if not to_delete:
            return

        for col_v in to_delete:
            col = collections_versioned[col_v]
            log.info(f"deleting {col} collection caused by retirement")
            self.client.drop_collection(col, timeout=self.milvus_timeout)

    def _update_alias(self, collection_name):
        try:
            self.client.create_alias(
                collection_name=collection_name,
                alias=self.collection_prefix,
                timeout=self.milvus_timeout,
            )
        except MilvusException:
            self.client.alter_alias(
                collection_name=collection_name,
                alias=self.collection_prefix,
                timeout=self.milvus_timeout,
            )

    def __construct_next_collection_name(self) -> str:
        previous_collections = self._get_collection_versions()
        if not previous_collections:
            return f"{self.collection_prefix}_v1"
        previous_version = max(previous_collections.keys())
        log.info(f"Found version v{previous_version}")
        return f"{self.collection_prefix}_v{previous_version + 1}"

    def __construct_last_collection_name(self) -> str:
        previous_collections = self._get_collection_versions()
        if not previous_collections or len(previous_collections) <= 1:
            raise NoPreviousCollection(f"Milvus does not contain a previous collection for {self.collection_prefix}")
        previous_version = sorted(previous_collections.keys())[-2]
        log.info(f"Found previous version v{previous_version}")
        return f"{self.collection_prefix}_v{previous_version}"

    def __construct_current_collection_name(self) -> str:
        previous_collections = self._get_collection_versions()
        if not previous_collections or len(previous_collections) < 1:
            raise NoPreviousCollection(f"Milvus does not contain a previous collection for {self.collection_prefix}")
        previous_version = sorted(previous_collections.keys())[-1]
        log.info(f"Found previous version v{previous_version}")
        return f"{self.collection_prefix}_v{previous_version}"

    def _get_collection_versions(self) -> dict[int, str]:
        previous_collections = self.client.list_collections(timeout=self.milvus_timeout)
        versioned_collections = {
            int(previous.split("_v")[-1]): previous for previous in previous_collections if self.collection_prefix in previous
        }
        return versioned_collections



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

from pandera.typing import Series

from wurzel.steps.embedding.data import EmbeddingMultiVectorResult, EmbeddingResult


class QdrantResult(EmbeddingResult):
    """Data Contract Proxy adding collection name to the PanderaDataframe."""

    text_tlsh_hash: Optional[Series[str]]
    collection: Series[str]
    id: Series[int]


class QdranttMultiVectorResult(EmbeddingMultiVectorResult):
    """Data Contract Proxy adding collection name to the PanderaDataframe."""

    text_tlsh_hash: Optional[Series[str]]
    collection: Series[str]
    id: Series[int]



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""contains Embedding Manager."""

from json.decoder import JSONDecodeError
from logging import getLogger
from re import Pattern as RegexPattern
from typing import Literal, Union

import requests
from langchain_core.embeddings import Embeddings
from pydantic import validate_call
from pydantic_core import Url

from wurzel.exceptions import (
    EmbeddingAPIException,
    EmbeddingException,
    UnrecoverableFatalException,
)

log = getLogger(__name__)


@validate_call
def _url_with_path(base: Url, path: str) -> Url:
    return Url.build(
        scheme=base.scheme,
        username=base.username,
        password=base.password,
        host=base.host,
        port=base.port,
        path=path,
        query=base.query,
        fragment=base.fragment,
    )


class HuggingFaceInferenceAPIEmbeddings(Embeddings):
    """Embed texts using the HuggingFace API.

    Requires a HuggingFace interface deployed as service
    """

    _timeout: int = 10
    embedding_url: Url
    info_url: Url
    _last_model: str
    _on_model_change: callable = None
    _normalize: bool = False

    @validate_call
    def __init__(self, url: Url, normalize: bool = False):
        self._normalize = normalize
        self.embedding_url = _url_with_path(url, "embed")
        self.info_url = _url_with_path(url, "info")
        self._last_model = None
        self._update_model_history(self.get_info()["model_id"])

    def _update_model_history(self, model: str) -> bool:
        """Updates internal model history if model name is new.

        Args:
            model (str): string name/ path of model

        Returns:
            bool: has the model changed

        """
        model_name = model.strip("/").split("/")[-1]
        log.info(f"Model history: name={model_name}")
        if self._last_model is None or model_name != self._last_model:
            self._last_model = model_name
            return True
        return False

    @validate_call
    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        """Get the embeddings for a list of texts."""
        return [self.embed_query(text) for text in texts]

    def __make_request(self, url: Url, json_body: dict, method: Union[Literal["post"], Literal["get"]]) -> dict:
        """Creates a request, tries to parse json.

        Args:
            url (str)
            json_body (dict): body to send

        Raises:
            EmbeddingAPIException: after timeout
            EmbeddingAPIException: invalid json
            EmbeddingAPIException: invalid statuscode

        Returns:
            dict: parsed response

        """
        try:
            response = requests.request(method, url, json=json_body, timeout=self._timeout)
        except (requests.exceptions.ConnectTimeout, requests.exceptions.Timeout) as err:
            raise EmbeddingAPIException(f"timed out after {self._timeout}") from err
        except requests.ConnectionError as err:
            raise EmbeddingAPIException("Connection Error") from err
        if response.status_code != 200:
            raise EmbeddingAPIException(f"failed, invalid status_code {response.status_code}")
        try:
            resp_dict = response.json()
        except JSONDecodeError as err:
            raise EmbeddingAPIException(f"failed due to invalid json {response.content}") from err

        return resp_dict

    def _request_embed_query(self, text: str) -> dict:
        return self.__make_request(
            self.embedding_url,
            {"inputs": text, "normalize": self._normalize},
            method="post",
        )

    @validate_call
    def embed_query(self, text: str) -> list[float]:
        """Compute query embeddings using a HuggingFace transformer model.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.

        """
        response = self._request_embed_query(text)
        try:
            value = response[0]
        except (KeyError, IndexError) as err:
            raise EmbeddingException(
                f"Response invalid Structure of received dict is incorrect: {response} should contain a list with one entry"
            ) from err
        return value

    def get_info(self):
        """Returns the infos of the model, described here:
        https://huggingface.github.io/text-embeddings-inference/#/.
        """
        response_model_key = "model_id"
        resp_dict = self.__make_request(self.info_url, None, method="get")
        if response_model_key not in resp_dict:
            raise EmbeddingException(f"Response invalid format {self.info_url} {resp_dict}requires {response_model_key}")
        return resp_dict


class PrefixedAPIEmbeddings(HuggingFaceInferenceAPIEmbeddings):
    """E5 and other models need a prefix within the input."""

    prefix: str = ""
    # Mapping function that should return None if it was not found
    prefix_mapping: dict[RegexPattern, str]

    @validate_call
    def __init__(self, url: Url, prefix_mapping: dict[RegexPattern, str]):
        super().__init__(url)
        self.prefix_mapping = prefix_mapping
        self.update_prefix()
        self._on_model_change = self.update_prefix

    @validate_call
    # overrides
    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        """Get the embeddings for a list of texts."""
        texts = [f"{self.prefix}{text}" for text in texts]
        embedding_response = super().embed_documents(texts)
        return embedding_response

    def update_prefix(self):
        """Updates prompt/ embedding prefix used internally in every embedded request.

        Raises:
            UnrecoverableFatalException: no prefix-definition for a model was found

        """
        for regex, prefix in self.prefix_mapping.items():
            if regex.search(self._last_model):
                self.prefix = prefix
                log.info(f"Using prefix={prefix}")
                return
        raise UnrecoverableFatalException(
            f"Tried to get prefix for {self._last_model}:" + f"No match found in {self.prefix_mapping.keys()}"
        )



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0
from wurzel.utils import HAS_MILVUS as _HAS_MILVUS

from .settings import MilvusSettings  # noqa: F401

if _HAS_MILVUS:
    from .step import MilvusConnectorStep  # noqa: F401



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from pydantic import Field

from wurzel.step.settings import Settings


class PrometheusSettings(Settings):
    """Used in PrometheusExecutor."""

    PROMETHEUS_GATEWAY: str = Field("localhost:9091", description="host:port of pushgateway")
    PROMETHEUS_JOB: str = Field("default-job-name", description="jobname for the prometheus counter")
    PROMETHEUS_DISABLE_CREATED_METRIC: bool = Field(True, description="disable *_created metrics")



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""Semantic Markdown Splitter."""

import re
from logging import getLogger
from typing import TYPE_CHECKING, Optional, TypedDict

import mdformat
import tiktoken
from mistletoe import Document as MisDocument
from mistletoe import block_token, markdown_renderer, span_token
from mistletoe.token import Token

from wurzel.datacontract import MarkdownDataContract
from wurzel.exceptions import MarkdownException
from wurzel.utils.to_markdown.html2md import MD_RENDER_LOCK

if TYPE_CHECKING:
    import spacy
LEVEL_MAPPING = {
    block_token.Heading: 0,  # actually 1-6
    block_token.List: 7,
    block_token.CodeFence: 8,
    block_token.Table: 9,
    block_token.Paragraph: 10,
    block_token.ListItem: 11,
    block_token.TableRow: 12,
    span_token.LineBreak: 13,
    block_token.ThematicBreak: 14,
}
LEVEL_UNDEFINED = 15
OPENAI_ENCODING = tiktoken.encoding_for_model("gpt-3.5-turbo")

log = getLogger(__name__)


class MetaDataDict(TypedDict):
    """Dict definition of metadata."""

    keywords: str
    url: str


class DocumentNode(TypedDict):
    """Dict definition of internal tree structure for splitting."""

    highest_level: int
    token_len: int
    text: str
    children: list["DocumentNode"]
    metadata: MetaDataDict


def _get_token_len(text: str) -> int:
    """Get OpenAI Token length.

    Args:
        text (str): Test to encode

    Returns:
        int: count of tokens

    """
    return len(OPENAI_ENCODING.encode(text))


def _is_all_children_same_level(children: list[DocumentNode]) -> bool:
    nbr_different_level = len({c["highest_level"] for c in children})
    return nbr_different_level == 1


def _get_children_sorted_by_level(
    children: list[DocumentNode],
) -> list[tuple[int, int]]:
    return sorted(
        [(i, c["highest_level"]) for i, c in enumerate(children)],
        key=lambda x: x[1],
        reverse=False,
    )


def _get_highest_index_of_children(children: list[DocumentNode]) -> int:
    sorted_by_level = _get_children_sorted_by_level(children)
    return sorted_by_level[0][0]


def _get_highest_level_of_children(children: list[DocumentNode]) -> int:
    sorted_by_level = _get_children_sorted_by_level(children)
    return sorted_by_level[0][1]


def _get_heading_text(token: block_token.Heading):
    """Get heading text from Mistletoe heading block token."""
    if token.content:
        return token.content
    if "children" in vars(token) and len(token.children) > 0:
        return " ".join([c.content for c in token.children if hasattr(c, "content")])
    return ""


def _is_standalone_a_heading(text):
    childs = MisDocument(text).children
    if len(childs) != 1:
        return False
    return isinstance(childs[0], block_token.Heading)


def _cut_to_tokenlen(text: str, token_len: int) -> str:
    """Cut Text to token length using OpenAI."""
    tokens_old = OPENAI_ENCODING.encode(text)
    if len(tokens_old) > token_len:
        tokens = tokens_old[:token_len]
        return OPENAI_ENCODING.decode(tokens)
    return OPENAI_ENCODING.decode(tokens_old)


def _format_markdown_docs(
    docs: list[MarkdownDataContract],
) -> list[MarkdownDataContract]:
    """Formats the Markdown Document by the standards."""
    return [
        MarkdownDataContract(
            md=mdformat.text(doc.md).strip(),
            url=doc.url,
            keywords=doc.keywords,
        )
        for doc in docs
    ]


class WurzelMarkdownRenderer(markdown_renderer.MarkdownRenderer):
    """Fix For markdown_renderer.MarkdownRenderer."""

    # pylint: disable=unused-argument, arguments-differ
    def render_table_cell(self, token: block_token.TableCell, max_line_length: int) -> str:
        """Renders the content of a table cell.

        Args:
            token (block_token.TableCell): The table cell token to render.
            max_line_length (int): The maximum allowed line length for the content.

        Returns:
            str: The rendered content of the table cell.

        """
        return self.render_inner(token)

    def render_table_row(self, token: block_token.TableRow, max_line_length: int) -> str:
        """Renders a table row from the given token.

        Args:
            token (block_token.TableRow): The table row token to be rendered.
            max_line_length (int): The maximum allowed line length for the rendered row.

        Returns:
            str: The rendered table row as a string.

        """
        return self.render_inner(token)


class SemanticSplitter:
    """Splitter implementation."""

    nlp: "spacy.language.Language"
    token_limit: int
    token_limit_buffer: int
    token_limit_min: int

    def __init__(
        self,
        token_limit: int = 256,
        token_limit_buffer: int = 32,
        token_limit_min: int = 64,
        spacy_model: str = "de_core_news_sm",
    ) -> None:
        """Initializes the SemanticSplitter class with specified token limits and a spaCy language model.

        Args:
            token_limit (int, optional): The maximum number of tokens allowed. Defaults to 256.
            token_limit_buffer (int, optional): The buffer size for token limit to allow flexibility. Defaults to 32.
            token_limit_min (int, optional): The minimum number of tokens required. Defaults to 64.
            spacy_model (str, optional): The name of the spaCy language model to load. Defaults to "de_core_news_sm".

        Raises:
            OSError: If the specified spaCy model cannot be loaded.

        """
        import spacy  # pylint: disable=import-outside-toplevel

        self.nlp = spacy.load(spacy_model)
        self.token_limit = token_limit
        self.token_limit_buffer = token_limit_buffer
        self.token_limit_min = token_limit_min

    def _is_short(self, text: str) -> bool:
        return _get_token_len(text) <= self.token_limit - self.token_limit_buffer

    def _is_table(self, doc: DocumentNode) -> bool:
        return doc["highest_level"] == LEVEL_MAPPING[block_token.Table]

    def _is_within_targetlen_w_buffer(self, text: str) -> bool:
        length = _get_token_len(text)
        return self.token_limit + self.token_limit_buffer >= length >= self.token_limit - self.token_limit_buffer

    def _render_doc(self, doc: MisDocument) -> str:
        """Render Mistletoe Markdown Document."""
        try:
            with MD_RENDER_LOCK, WurzelMarkdownRenderer() as renderer:
                return renderer.render(doc)  # type: ignore[no-any-return]
        except Exception as e:
            raise MarkdownException(e) from e

    def _get_custom_level(self, block: block_token.BlockToken) -> int:
        """Get the hierarchical level for a mistletoe node."""
        if isinstance(block, block_token.Heading):
            return int(block.level)
        return LEVEL_MAPPING.get(type(block), LEVEL_UNDEFINED)

    def _merge_children(self, children: list[DocumentNode]) -> MisDocument:
        """Create a document out of a list of children."""
        new_doc = MisDocument([])
        # If all children a span tokens add them to a paragraph
        # because problems otherwise
        if all(isinstance(c, span_token.SpanToken) for c in children):
            para = block_token.Paragraph([])
            para.children = children
            new_doc.children += [para]
        else:
            new_doc.children += children
        return new_doc

    def _find_highest_level(self, children: list[DocumentNode], min_level: int = 0) -> tuple[int, Optional[Token], Optional[DocumentNode]]:
        """Among a list of children nodes find the one with the highest level.

        Return a tuple of that level, level node type and that child.
        """

        def is_any_children(children: list[DocumentNode], block_type: type[block_token.BlockToken]):
            """Check if any Mistletoe Node (child) is of specific type."""
            for child in children:
                if isinstance(child, block_type):
                    return True
            return False

        highest_level: int = LEVEL_UNDEFINED
        highest_type: Optional[type[Token]] = None
        highest_element: Optional[DocumentNode] = None

        if children is None:
            return (LEVEL_UNDEFINED, None, None)
        # Checked by higher level function
        for child in children:
            if isinstance(child, block_token.Heading) and child.level > min_level:
                if child.level < highest_level:
                    highest_level = int(child.level)
                    highest_type = block_token.Heading
                    highest_element = child

        # block_token.ThematicBreak
        # If level is not set by the Heading
        for block_type in [
            block_token.List,
            block_token.CodeFence,
            block_token.Table,
            block_token.Paragraph,
            block_token.ListItem,
            block_token.TableRow,
            span_token.LineBreak,  # Add table
            block_token.ThematicBreak,
        ]:
            if (highest_level == LEVEL_UNDEFINED) and is_any_children(children, block_type) and (LEVEL_MAPPING[block_type] > min_level):
                highest_level = LEVEL_MAPPING[block_type]
                highest_type = block_type
                highest_element = [child for child in children if isinstance(child, block_type)][0]
                break

        return highest_level, highest_type, highest_element

    def _split_children(self, children: list[MisDocument], min_level: int = 0) -> list[MisDocument]:
        """Split a list of children in the most semantic way and return a list of Documents with them merged."""
        if len(children) == 1:
            if "children" in vars(children[0]) or "_children" in vars(children[0]):
                return self._split_children(children[0].children)
            return [children[0]]
        highest_level, highest_type, _ = self._find_highest_level(children, min_level)

        # No higher splitter found
        if highest_level == LEVEL_UNDEFINED:
            return children
        assert highest_type
        # Find point to split the list of children
        split_points: list[int] = []
        for i, child in enumerate(children):
            if isinstance(child, highest_type):
                if highest_level == self._get_custom_level(child):
                    split_points.append(i)

        # There must be at least one split if highest_level < LEVEL_UNDEFINED
        # Highest splitter is first element
        if (split_points[0] == 0) and len(split_points) == 1:
            # Remove leading line breaks
            if isinstance(children[0], span_token.LineBreak):
                children = children[1:]
            # run again with lower level requirement
            return self._split_children(children, highest_level)

        # Combine Children to Documents
        prev_i = 0
        return_docs = []
        for i in split_points:
            if children[prev_i:i]:
                new_doc = self._merge_children(children[prev_i:i])
                return_docs.append(new_doc)
            prev_i = i
        new_doc = self._merge_children(children[split_points[-1] :])
        return_docs.append(new_doc)
        return return_docs

    def text_sentences(self, text):
        """Split a text into sentences using a NLP model.

        This does not use a Regex based approach on purpose as they break with
        punctuation very easily see: https://stackoverflow.com/a/61254146
        """
        return [sentence_span.text for sentence_span in self.nlp(text).sents]

    def _markdown_hierarchy_parser(self, text: str, metadata: MetaDataDict, max_depth: int = 30) -> DocumentNode:
        """Splits a Markdown string into a semantic Markdown based hierarchy.

        Given a Markdown string it hierarchically splits that text using
        the semantic information from the Markdown document until
        all final leaf nodes are below the global token limit.

        max_depth decreases with every recursive call of the function
        controlling the maximum depth of the hierarchy.
        """
        md_doc = MisDocument(text)
        highest_level, _, _ = self._find_highest_level(md_doc.children)
        token_len = _get_token_len(text)

        # Reached max recursion depth
        if max_depth == 0:
            log.warning("maximal markdown recursion reached")
            return DocumentNode(
                highest_level=highest_level,
                token_len=token_len,
                text=text,
                metadata=metadata,
                children=[],
            )

        # Do further hierarchy parsing
        splits: MisDocument = self._split_children(md_doc.children)

        def has_node_a_known_level(x):
            return any(isinstance(x.children[0] if isinstance(x, MisDocument) else x, cl) for cl in LEVEL_MAPPING)

        no_child_has_a_known_level = not any(has_node_a_known_level(x) for x in splits)
        # No child has a known level this means we only have sentences and no more semantic information
        if no_child_has_a_known_level:
            return DocumentNode(
                highest_level=highest_level,
                token_len=token_len,
                text=text,
                metadata=metadata,
                children=[
                    DocumentNode(
                        highest_level=highest_level,
                        token_len=_get_token_len(sent),
                        text=sent,
                        metadata=metadata,
                        children=[],
                    )
                    for sent in self.text_sentences(text)
                ],
            )

        # Further split the child nodes until we reach the token limit for each
        children: list[DocumentNode] = []
        for s in splits:
            if not hasattr(s, "children"):
                continue
            md_child = self._render_doc(s)
            highest_level_child, _, _ = self._find_highest_level(s.children)
            children.append(
                DocumentNode(
                    highest_level=highest_level_child,
                    text=md_child,
                    token_len=_get_token_len(md_child),
                    children=[self._markdown_hierarchy_parser(md_child, metadata, max_depth - 1)],
                    metadata=metadata,
                )
            )
        return DocumentNode(
            highest_level=highest_level,
            token_len=token_len,
            text=text,
            children=children,
            metadata=metadata,
        )

    # unused
    def _split_by_sentence(self, text: str) -> list[str]:
        """Sometimes _split_children does not find children leafs with are smaller then TOKEN_LIMIT.

        Thus we need to split by sentence.
        """
        token_limit = self.token_limit
        token_buffer = self.token_limit_buffer
        lenth = _get_token_len(text)
        needed_splits = lenth // token_limit
        if not needed_splits:
            return [text]
        sentences = [(_get_token_len(sent), f"{sent}. ") for sent in re.split(r"\.(?=\s|\\n)", text) if sent.strip()]
        chunks: list[str] = []
        chunk = ""
        chunk_len = 0
        for size, sent in sentences:
            if size > token_limit + token_buffer:  # single big sentence
                if chunk:  # then add last
                    chunks.append(chunk)
                    chunk = ""
                    chunk_len = 0
                chunks.append(_cut_to_tokenlen(sent, token_limit))  # cut this
                # Last piece of sentence is discarded
                continue
            if chunk_len + size > token_limit + token_buffer:  # with next to big
                chunks.append(chunk)
                chunk = ""
                chunk_len = 0
            chunk += sent
            chunk_len += size
            if token_limit + token_buffer >= chunk_len >= token_limit - token_buffer:  # together they fit
                chunks.append(chunk)

                chunk = ""
                chunk_len = 0
        if chunk:
            chunks.append(_cut_to_tokenlen(chunk, token_limit))
        # chunks = [
        #    (chunk.replace("\n").strip() if not "#" else chunk.strip()) for chunk in chunks
        # ] This was broken
        chunks = [
            (_cut_to_tokenlen(chunk, token_limit) if not self._is_within_targetlen_w_buffer(chunk) or self._is_short(chunk) else chunk)
            for chunk in chunks
        ]
        for chunk in chunks:
            assert self._is_within_targetlen_w_buffer(chunk) or self._is_short(chunk)
        return chunks

    # pylint: disable-next=too-many-positional-arguments
    def _handle_parsing_of_children(
        self,
        doc: DocumentNode,
        child: DocumentNode,
        text_w_prev_child: str,
        remaining_snipped: str,
        recursive_depth: int,
    ) -> tuple[str, list[MarkdownDataContract]]:
        """Handle the parsing of child nodes during hierarchical parsing.

        This method is used to process child nodes of a document node and determine
        how their text content should be handled based on specific conditions. It
        helps manage recursion depth and ensures the text is split into manageable
        Markdown data contracts.

            doc (DocumentNode): The parent document node containing metadata and text.
            child (DocumentNode): The child document node to be processed.
            text_w_prev_child (str): The text content combined with the previous child node.
            remaining_snipped (str): The remaining text snippet to be processed.
            recursive_depth (int): Tracks the current recursion depth during hierarchical parsing.

            tuple[str, list[MarkdownDataContract]]:
                - A string representing the remaining unprocessed text snippet.
                - A list of MarkdownDataContract objects containing processed Markdown data.

        """
        return_doc = []
        if self._is_short(text_w_prev_child):
            remaining_snipped = text_w_prev_child
        elif self._is_within_targetlen_w_buffer(text_w_prev_child):
            child["text"] = text_w_prev_child
            return_doc += [
                MarkdownDataContract(
                    md=_cut_to_tokenlen(child["text"], self.token_limit),
                    url=child["metadata"]["url"],
                    keywords=child["metadata"]["keywords"],
                )
            ]
            remaining_snipped = ""
        else:
            if not _is_standalone_a_heading(remaining_snipped):
                if _get_token_len(remaining_snipped) >= self.token_limit_min:
                    return_doc.append(
                        MarkdownDataContract(
                            md=remaining_snipped,
                            keywords=doc["metadata"]["keywords"],
                            url=doc["metadata"]["url"],
                        )
                    )
                remaining_snipped = ""
                if self._is_within_targetlen_w_buffer(child["text"]):
                    return_doc.append(self._md_data_from_dict_cut(child))
                else:
                    return_doc += self._parse_hierarchical(child, recursive_depth + 1)
            else:
                temp_docs = self._parse_hierarchical(child, recursive_depth + 1)
                return_doc += [
                    MarkdownDataContract(
                        md=remaining_snipped + "\n\n" + d.md,
                        keywords=d.keywords,
                        url=d.url,
                    )
                    for d in temp_docs
                ]
        return remaining_snipped, return_doc

    def _md_data_from_dict_cut(self, doc):
        return MarkdownDataContract(
            md=_cut_to_tokenlen(doc["text"], self.token_limit),
            url=doc["metadata"]["url"],
            keywords=doc["metadata"]["keywords"],
        )

    def _parse_hierarchical(
        self,
        doc: DocumentNode,
        recursive_depth: int = 1,
    ) -> list[MarkdownDataContract]:
        if _get_token_len(doc["text"]) <= self.token_limit_min:
            if recursive_depth == 1:
                return [self._md_data_from_dict_cut(doc)]
            log.warning("document to short", extra=doc)
            return []
        if self._is_within_targetlen_w_buffer(doc["text"]):
            return [self._md_data_from_dict_cut(doc)]
        if "children" not in doc.keys():
            log.warning(
                "no remaining children. still to big -> sentence split by dot",
                extra=doc,
            )
            return [self._md_data_from_dict_cut(doc)]
        if self._is_table(doc):
            log.warning(
                "found table, that should have been split, cutting off",
                extra=doc["metadata"],
            )
            return [self._md_data_from_dict_cut(doc)]
        if len(doc["children"]) == 0:
            log.warning(
                "no remaining children. still to big -> Cut by tokenlen",
                extra=doc,
            )
            return [self._md_data_from_dict_cut(doc)]
        if len(doc["children"]) == 1:
            return self._parse_hierarchical(doc["children"][0])

        children = doc["children"]
        return_doc: list[MarkdownDataContract] = []

        # If we don't have a proper hierarchy, handle the elements before top item separately
        if not _is_all_children_same_level(children):
            idx_highest = _get_highest_index_of_children(children)
            highest_child_is_heading = _get_highest_level_of_children(children) <= 6
            first_child_is_highest = idx_highest == 0
            if not first_child_is_highest and highest_child_is_heading:
                text_until_highest_child = "\n".join([c["text"] for c in children[:idx_highest]])
                token_len = _get_token_len(text_until_highest_child)
                max_level = max(c["highest_level"] for c in children[:idx_highest])
                new_doc = DocumentNode(
                    highest_level=max_level,
                    token_len=token_len,
                    text=text_until_highest_child,
                    children=children[:idx_highest],
                    metadata=doc["metadata"],
                )
                return_doc += self._parse_hierarchical(new_doc)
                children = children[idx_highest:]

        remaining_snipped = ""
        for child in children:
            child["metadata"] = doc["metadata"]  # inherit metadata downwards
            text_w_prev_child = "\n".join([remaining_snipped, child["text"]]).strip()
            remaining_snipped, returned_docs = self._handle_parsing_of_children(
                doc, child, text_w_prev_child, remaining_snipped, recursive_depth
            )
            return_doc += returned_docs

        # add potential short remaining spillovers
        if _get_token_len(remaining_snipped) >= self.token_limit_min:
            return_doc += [
                MarkdownDataContract(
                    md=_cut_to_tokenlen(remaining_snipped, self.token_limit),
                    url=doc["metadata"]["url"],
                    keywords=doc["metadata"]["keywords"],
                )
            ]
        return return_doc

    def _adopt_splitted_list_to_use_highest_prev_header(
        self,
        docs: list[MarkdownDataContract],
    ) -> list[MarkdownDataContract]:
        """Function to improve the semantic meaning of the Markdown document by reattaching a parent heading.

        Does not yet respect the token limit, however headings usually have little impact
        """
        highest_header_until_now = {i + 1: "" for i in range(6)}
        for doc in docs:
            text = doc.md.strip()
            md_doc = MisDocument(text)
            # Create Header
            highest_level, highest_type, highest_child = self._find_highest_level(md_doc.children)
            if str(highest_type) == str(block_token.Heading):
                # Discuss this
                # highest_header_until_now[highest_level] = _get_heading_text(highest_child)
                highest_header_until_now[highest_level] = self._render_doc(highest_child).lstrip(" #")

            ordered_headers = {level: text for level, text in highest_header_until_now.items() if text}
            docwide_highest_level = min(ordered_headers.keys()) if ordered_headers else 10
            new_header = " - ".join([text for level, text in ordered_headers.items() if level < highest_level])
            if new_header:
                new_header = "# " + new_header
            # Filter Doc
            document_is_just_single_header = text.strip().startswith("#") and "\n" not in text.strip()
            if document_is_just_single_header:
                continue
            # TODOo check token limit or limit header lenght
            # The higher heading the lower its level
            doc_has_lower_heading = docwide_highest_level < highest_level
            new_doc = text.strip() if not doc_has_lower_heading else (new_header + "\n\n" + doc.md).strip()
            doc.md = new_doc

        return docs

    def split_markdown_document(self, doc: MarkdownDataContract) -> list[MarkdownDataContract]:
        """Split a Markdown Document into Snippets."""
        metadata = MetaDataDict(url=doc.url, keywords=doc.keywords)
        doc_hierarchy: DocumentNode = self._markdown_hierarchy_parser(doc.md, metadata)
        doc_snippets: list[MarkdownDataContract] = self._parse_hierarchical(doc_hierarchy)
        improved_snippets: list[MarkdownDataContract] = self._adopt_splitted_list_to_use_highest_prev_header(doc_snippets)
        formatted_snippets: list[MarkdownDataContract] = _format_markdown_docs(improved_snippets)
        return formatted_snippets



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import json

from pydantic import Field, validator

from wurzel.step.settings import Settings


# pylint: disable=duplicate-code
class MilvusSettings(Settings):
    """MilvusSettings is a configuration class for managing settings related to MilvusDB.

    Attributes:
        HOST (str): The hostname or IP address of the Milvus server. Defaults to "localhost".
        PORT (int): The port number for the Milvus server. Must be between 1 and 65535. Defaults to 19530.
        COLLECTION (str): The name of the collection in MilvusDB.
        COLLECTION_HISTORY_LEN (int): The length of the collection history. Defaults to 10.
        SEARCH_PARAMS (dict): Parameters for search operations in MilvusDB. Defaults to {"metric_type": "IP", "params": {}}.
        INDEX_PARAMS (dict): Parameters for indexing operations in MilvusDB. Defaults to {"index_type": "FLAT",
                                "field_name": "vector", "metric_type": "IP", "params": {}}.
        USER (str): The username for authentication with MilvusDB.
        PASSWORD (str): The password for authentication with MilvusDB.
        SECURED (bool): Indicates whether the connection to MilvusDB is secured. Defaults to False.

    Methods:
        parse_json(cls, v): Validates and parses JSON strings into Python objects for SEARCH_PARAMS and INDEX_PARAMS.

    Args:
                v (str or dict): The value to validate and parse.

    Returns:
                dict: The parsed dictionary if the input is a JSON string, or the original value if it is already a dictionary.

    """

    HOST: str = "localhost"
    PORT: int = Field(19530, gt=0, le=65535)
    COLLECTION: str
    COLLECTION_HISTORY_LEN: int = 10
    SEARCH_PARAMS: dict = {"metric_type": "IP", "params": {}}
    INDEX_PARAMS: dict = {
        "index_type": "FLAT",
        "field_name": "vector",
        "metric_type": "IP",
        "params": {},
    }
    USER: str
    PASSWORD: str
    SECURED: bool = False

    @validator("SEARCH_PARAMS", "INDEX_PARAMS", pre=True)
    @classmethod
    # pylint: disable-next=R0801
    def parse_json(cls, v):
        """Validation for json."""
        if isinstance(v, str):
            return json.loads(v)
        return v



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path

from wurzel.adapters.dvc_adapter import Backend
from wurzel.step.typed_step import TypedStep
from wurzel.step_executor.base_executor import BaseStepExecutor
from wurzel.utils import create_model


def main(step: TypedStep, data_dir: Path, backend: Backend) -> str:
    """Generates the yaml for the given backend."""
    adapter: Backend = backend(data_dir)
    # validate the envs of the steps
    create_model(
        list(step.traverse()),
        allow_extra_fields=BaseStepExecutor.is_allow_extra_settings(),
    )()

    return adapter.generate_yaml(step)



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""consists of DVCSteps to embedd files and save them as for example as csv."""

# Standard library imports
from logging import getLogger

from joblib import Parallel, delayed
from pydantic import Field

from wurzel.datacontract import MarkdownDataContract
from wurzel.exceptions import MarkdownException, SplittException
from wurzel.step import Settings, TypedStep
from wurzel.utils.semantic_splitter import SemanticSplitter

# Local application/library specific imports


class SplitterSettings(Settings):
    """Anything Embedding-related."""

    BATCH_SIZE: int = Field(100, gt=0)
    NUM_THREADS: int = Field(4, gt=1)
    TOKEN_COUNT_MIN: int = Field(64, gt=0)
    TOKEN_COUNT_MAX: int = Field(1024, gt=1)
    TOKEN_COUNT_BUFFER: int = Field(32, gt=0)


log = getLogger(__name__)


class SimpleSplitterStep(TypedStep[SplitterSettings, list[MarkdownDataContract], list[MarkdownDataContract]]):
    """SimpleSplitterStep to split Markdown Documents rundimentory in medium size chunks."""

    def __init__(self) -> None:
        super().__init__()
        self.splitter = SemanticSplitter(
            token_limit=self.settings.TOKEN_COUNT_MAX,
            token_limit_buffer=self.settings.TOKEN_COUNT_BUFFER,
            token_limit_min=self.settings.TOKEN_COUNT_MIN,
        )

    def run(self, inpt: list[MarkdownDataContract]) -> list[MarkdownDataContract]:
        """Executes the split step by processing input markdown files and generating smaller split markdown files and preservs the headline.

        Args:
            inpt (list[MarkdownDataContract]): A list of MarkdownDataContract objects representing
                                               the input markdown files to be processed.

        Returns:
            list[MarkdownDataContract]: A list of MarkdownDataContract objects representing
                                         the smaller split markdown files.

        The method splits the input markdown files into batches based on the configured batch size
        (`self.settings.BATCH_SIZE`). Each batch is processed in parallel using threading, and the
        results are flattened into a single list of split markdown files.

        """

        def _batchify(data: list, size: int) -> list[list]:
            return [data[i : i + size] for i in range(0, len(data), size)]

        batches = _batchify(inpt, self.settings.BATCH_SIZE)

        # Run each batch in parallel using threading
        results = Parallel(n_jobs=self.settings.NUM_THREADS, prefer="threads")(delayed(self._split_markdown)(batch) for batch in batches)

        # Flatten the list of lists
        return [item for sublist in results for item in sublist]

    def _split_markdown(self, markdowns: list[MarkdownDataContract]) -> list[MarkdownDataContract]:
        """Creates data rows from a batch of markdown texts by splitting them and counting tokens."""
        rows = []
        skipped = 0
        for s in markdowns:
            try:
                rows.extend(self.splitter.split_markdown_document(s))
            except MarkdownException as err:
                log.warning(
                    "skipped dokument ",
                    extra={"reason": err.__class__.__name__, "doc": s},
                )
                skipped += 1
        if skipped == len(markdowns):
            raise SplittException("all Documents got skipped during splitting")
        if skipped:
            log.error(f"{(skipped / len(markdowns)) * 100}% got skipped")
        return rows



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""consists of DVCSteps to embedd files and save them as for example as csv."""

# Standard library imports
from logging import getLogger

import pandas as pd

from wurzel.datacontract import MarkdownDataContract
from wurzel.step import Settings, TypedStep

# Local application/library specific imports


log = getLogger(__name__)


class DropStettings(Settings):
    """specify DROP_BY_FIELDS to field."""

    DROP_BY_FIELDS: list[str] = ["md"]


class DropDuplicationStep(TypedStep[DropStettings, list[MarkdownDataContract], list[MarkdownDataContract]]):
    """SimpleSplitterStep to split Markdown Documents rundimentory in medium size chunks."""

    def run(self, inpt: list[MarkdownDataContract]) -> list[MarkdownDataContract]:
        """Executes the split step by processing input markdown files, generating smaller splitted markdown files,
        by preserving the headline.
        """
        if self.settings.DROP_BY_FIELDS == ["*"]:
            self.settings.DROP_BY_FIELDS = None
        df = pd.DataFrame(i.model_dump() for i in inpt)
        if not df.duplicated(self.settings.DROP_BY_FIELDS).any():
            return inpt

        filtered = df.drop_duplicates(self.settings.DROP_BY_FIELDS)
        log.warning(
            "Removed duplicates",
            extra={
                "percentage": len(filtered) / len(df),
                "before": len(df),
                "after": len(filtered),
                "by": str(self.settings.DROP_BY_FIELDS),
            },
        )
        dumped = filtered.to_dict(orient="records")
        return [MarkdownDataContract.model_construct(**f) for f in dumped]



# SPDX-FileCopyrightText: 2024 Deutsche Telekom AG
#
# SPDX-License-Identifier: CC0-1.0

"""Specific docling settings."""

from docling.datamodel.base_models import InputFormat

from wurzel.step import Settings


class DoclingSettings(Settings):
    """DoclingSettings is a configuration class that inherits from the base `Settings` class.
    It provides customizable settings for document processing.

    Attributes:
        FORCE_FULL_PAGE_OCR (bool): A flag to enforce full-page OCR processing. Defaults to True.
        FORMATS (list[InputFormat]): A list of supported input formats for document processing.
            Supported formats include:
            - "docx"
            - "asciidoc"
            - "pptx"
            - "html"
            - "image"
            - "pdf"
            - "md"
            - "csv"
            - "xlsx"
            - "xml_uspto"
            - "xml_jats"
            - "json_docling"
        URLS (list[str]): A list of URLs for additional configuration or resources. Defaults to an empty list.

    """

    FORCE_FULL_PAGE_OCR: bool = True
    FORMATS: list[InputFormat] = [
        "docx",
        "asciidoc",
        "pptx",
        "html",
        "image",
        "pdf",
        "md",
        "csv",
        "xlsx",
        "xml_uspto",
        "xml_jats",
        "json_docling",
    ]
    URLS: list[str] = []
    DEFAULT_KEYWORD: str = ""



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from wurzel.utils import HAS_QDRANT as _HAS_QDRANT

from .data import QdrantResult  # noqa: F401

if _HAS_QDRANT:
    from .step import QdrantConnectorStep  # noqa: F401
    from .step_multi_vector import QdrantConnectorMultiVectorStep  # noqa: F401

else:
    pass



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from .docling import *  # noqa: F403 Allow importing Step classes from docling
from .embedding import *  # noqa: F403 Allow importing Step classes
from .embedding import EmbeddingStep  # noqa: F401
from .manual_markdown import ManualMarkdownStep  # noqa: F401
from .milvus import *  # noqa: F403 Allow importing Step classes
from .qdrant import *  # noqa: F403 Allow importing Step classes



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import json
from inspect import getfile
from types import NoneType

from pydantic_core import PydanticUndefined

from wurzel.step import Settings, TypedStep
from wurzel.utils import WZ


def main(step: type[TypedStep], gen_env=False):
    """Execute."""
    ins = WZ(step)
    set_cls: Settings = ins.settings_class
    env_prefix = step.__name__.upper()
    data = {
        "Name": step.__name__,
        "Input": "None" if ins.input_model_class == NoneType else ins.input_model_class,
        "Output": ins.output_model_type,
        "settings": {
            "env_prefix": env_prefix,
            "fields": {k: str(v) for k, v in set_cls.model_fields.items()},
        },
    }
    if gen_env:
        setts = {True: [], False: []}
        for name, info in set_cls.model_fields.items():
            default = info.get_default(call_default_factory=True)
            default = "" if default == PydanticUndefined or default is None else default
            setts[info.is_required()].append(f"{env_prefix}__{name}={default}")
        print(f"# Env for {step.__name__} -> {getfile(step)}")
        print("# Required")
        print("\n".join(setts[True]))
        print("# Optional")
        print("\n".join(setts[False]))
    else:
        print(json.dumps(data, indent="  ", default=str))
